"""
This provides a card extension for the website.
"""

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx.application import Sphinx

__all__ = ["Card", "_Card", "depart_card_node", "visit_card_node"]


class _Card(nodes.General, nodes.Element):
    pass


def visit_card_node(self, node) -> None:
    """
    Prepare the card node for rendering.
    """
    title = node.get("title", "")
    key = title or node["github"]
    key = key.lower().replace(" ", "-")
    title = f"<h4>{title}</h4>" if len(title) > 0 else ""
    col_extra_class = "column-half" if title else ""
    img_src = node.get("img_name")
    # If there is no "img_name" given, we fallback to using the github avatar
    # if a user handle is provided. If so, the image provided is actually the sunpy icon
    if img_src.endswith("sunpy_icon.svg"):
        # The / works on RTD on the website repo, it is broken in this repo and on circleci
        img_src = f"https://github.com/{node['github']}.png" if node.get("github") else "/_static/img/sunpy_icon.svg"

    body = f"""<div class="column {col_extra_class}">
                {title}
                <div class="card">
                <img class="dark-light" src="{img_src}" alt="{node["name"]}">
                <p>{node["name"]}</p>
                <p><button type="button" class="btn btn-sunpy btn-sunpy1 stretched-link" data-bs-toggle="modal" data-bs-target="#{key}">More Info</button></p>
                <div class="modal fade" id="{key}" tabindex=-1>
                    <div class="modal-dialog modal-lg">
                        <div class="modal-content">
                            <div class="modal-header">
                                <h4 class="modal-title center">{node["name"]}</h4>
                                <button type="button" class="btn-close" data-bs-dismiss="modal" aria-label="Close"></button>
                            </div>
                            <div class="modal-body">
    """
    self.body.append(body)


def depart_card_node(self, node) -> None:
    """
    Finalize the card node after rendering.
    """
    start_date = f"<p>Start Date: {node['date']}</p>" if node.get("date") else ""
    body = f"""
                                <p>Affiliation: <a href="{node["aff_link"]}">{node["aff_name"]}</a></p>
                                <p>GitHub: <a href="https://github.com/{node["github"]}">{node["github"]}</a></p>
                                {start_date}
                            </div>
                        </div>
                    </div>
                </div>
            </div></div>"""
    self.body.append(body)


class Card(Directive):
    """
    A custom directive for a card.
    """

    has_content = True
    required_arguments = 1
    optional_arguments = 6
    option_spec = {  # NOQA: RUF012
        "img_name": directives.unchanged,
        "title": directives.unchanged,
        "github": directives.unchanged,
        "aff_name": directives.unchanged,
        "aff_link": directives.unchanged,
        "date": directives.unchanged,
        "desc": directives.unchanged,
    }

    def run(self):
        """
        Run the directive.
        """
        title = self.options.get("title") if "title" in self.options else ""
        img_name = self.options.get("img_name") if "img_name" in self.options else "sunpy_icon.svg"
        github = self.options.get("github") if "github" in self.options else ""
        aff_name = self.options.get("aff_name") if "aff_name" in self.options else ""
        aff_link = self.options.get("aff_link") if "aff_link" in self.options else ""
        date = self.options.get("date") if "date" in self.options else ""
        desc = self.options.get("desc") if "desc" in self.options else "N/A"
        name = " ".join(self.arguments)
        out = _Card(
            name=name,
            img_name=img_name,
            title=title,
            github=github,
            aff_name=aff_name,
            aff_link=aff_link,
            date=date,
            desc=desc,
        )
        self.state.nested_parse(self.content, 0, out)
        return [out]


def setup(app: Sphinx):
    app.add_css_file("css/sunpy_cards.css", priority=600)
    app.add_directive("custom-card", Card)
    app.add_node(_Card, html=(visit_card_node, depart_card_node))
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""
Configuration file for the Sphinx documentation builder.
"""

import datetime
import os
import sys
from pathlib import Path

from sphinx_gallery.sorting import ExplicitOrder

from sunpy_sphinx_theme import SVG_ICON, _sunpy_static_path

# Add the test package to the path so we can import it for automodapi
sys.path.append(Path().absolute().as_posix())
# This serves as a check
import test_package  # NOQA: F401

project = "sunpy-sphinx-theme test docs"
author = "The SunPy Community"
copyright = f"{datetime.datetime.now(datetime.timezone.utc).year}, {author}"  # NOQA: A001
extensions = [
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "hoverxref.extension",
    "matplotlib.sphinxext.plot_directive",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_gallery.gen_gallery",
    "sphinx_togglebutton",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sunpy_sphinx_theme.cards",
]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
source_suffix = ".rst"
master_doc = "index"
default_role = "obj"
napoleon_use_rtype = False
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/python3.inv"),
    ),
    "numpy": (
        "https://numpy.org/doc/stable/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/numpy.inv"),
    ),
    "scipy": (
        "https://docs.scipy.org/doc/scipy/reference/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/scipy.inv"),
    ),
    "matplotlib": (
        "https://matplotlib.org/",
        (None, "http://www.astropy.org/astropy-data/intersphinx/matplotlib.inv"),
    ),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/latest/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "skimage": ("https://scikit-image.org/docs/stable/", None),
    "drms": ("https://docs.sunpy.org/projects/drms/en/stable/", None),
    "parfive": ("https://parfive.readthedocs.io/en/latest/", None),
    "reproject": ("https://reproject.readthedocs.io/en/stable/", None),
}
html_theme = "sunpy"
html_static_path = [str(_sunpy_static_path), "_static"]
html_extra_path = ["_static/img"]
html_theme_options = {
    "footer_links": [
        ("Google", "https://google.com", 3),
        ("DDG", "https://duckduckgo.com", 3),
    ],
    "external_links": [
        {"name": "Python", "url": "https://www.python.org/"},
    ],
    "goatcounter_non_domain_endpoint": "https://sunpy-testing.goatcounter.com/count",
}
graphviz_output_format = "svg"
graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]
sphinx_gallery_conf = {
    "filename_pattern": "^((?!skip_).)*$",
    "examples_dirs": str(Path("..") / Path("examples")),
    "subsection_order": ExplicitOrder(
        [
            "../examples/section",
        ]
    ),
    "within_subsection_order": "ExampleTitleSortKey",
    "gallery_dirs": str(Path("generated") / Path("gallery")),
    "default_thumb_file": SVG_ICON,
    "abort_on_example_error": False,
    "plot_gallery": "True",
    "remove_config_comments": True,
}
if os.environ.get("READTHEDOCS"):
    hoverxref_api_host = "https://readthedocs.org"

    if os.environ.get("PROXIED_API_ENDPOINT"):
        # Use the proxied API endpoint
        # A RTD thing to avoid a CSRF block when docs are using a custom domain
        hoverxref_api_host = "/_"

hoverxref_auto_ref = False
hoverxref_domains = ["py"]
hoverxref_mathjax = True
hoverxref_modal_hover_delay = 500
hoverxref_tooltip_maxwidth = 600  # RTD main window is 696px
hoverxref_intersphinx = list(intersphinx_mapping.keys())
hoverxref_role_types = {
    # Roles within the py domain
    "attr": "tooltip",
    "class": "tooltip",
    "const": "tooltip",
    "data": "tooltip",
    "exc": "tooltip",
    "func": "tooltip",
    "meth": "tooltip",
    "mod": "tooltip",
    "obj": "tooltip",
    # Roles within the std domain
    "confval": "tooltip",
    "hoverxref": "tooltip",
    "ref": "tooltip",  # Would be used by hoverxref_auto_ref if we set it to True
    "term": "tooltip",
}
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True



"""
=================
<Verbing> a thing
=================

How to <verb> <active tense> <does something>.

The example uses <packages> to <do something> and <other package> to <do other
thing>. Include links to referenced packages like this: `sunpy.io.fits` to
show the sunpy.io.fits or like this `~sunpy.io.fits` to show just 'fits'
"""

###########################################################################
# If you have multiple plots in your example, Sphinx will use the first
# plot for the thumbnail by default. You can set which plot is used as the
# thumbnail by adding the following comment to your file.
# ``#sphinx_gallery_thumbnail_number = X``
# where ``X`` is the figure number you want (from 1).

import matplotlib.pyplot as plt
import numpy as np

##############################################################################
# This code block is executed, although it produces no output. Lines starting
# with a simple hash are code comment and get treated as part of the code
# block. To include this new comment string we started the new block with a
# long line of hashes.
#
# The sphinx-gallery parser will assume everything after this splitter and that
# continues to start with a **comment hash and space** (respecting code style)
# is text that has to be rendered in
# html format. Keep in mind to always keep your comments always together by
# comment hashes. That means to break a paragraph you still need to comment
# that line break.
#
# In this example the next block of code produces some plotable data. Code is
# executed, figure is saved and then code is presented next, followed by the
# inlined figure.

x = np.linspace(-np.pi, np.pi, 300)
xx, yy = np.meshgrid(x, x)
z = np.cos(xx) + np.cos(yy)

plt.figure()
plt.imshow(z)
plt.colorbar()
plt.xlabel("$x$")
plt.ylabel("$y$")

###########################################################################
# Again it is possible to continue the discussion with a new Python string. This
# time to introduce the next code block generates 2 separate figures.

plt.figure()
plt.imshow(z, cmap=plt.get_cmap("hot"))
plt.figure()
plt.imshow(z, cmap=plt.get_cmap("Spectral"), interpolation="none")

##########################################################################
# There's some subtle differences between rendered html rendered comment
# strings and code comment strings which I'll demonstrate below.
#
# Comments in comment blocks remain nested in the text.


def dummy() -> None:
    """
    Dummy function to make sure docstrings don't get rendered as text.
    """
    print("This is a dummy function")


# Code comments not preceded by the hash splitter are left in code blocks.
string = """Triple-quoted string which tries to break parser but doesn't."""

############################################################################
# Output of the script is captured:

print("Some output from Python")

############################################################################
# Finally, I'll call ``show`` at the end just so someone running the Python
# code directly will see the plots; this is not necessary for creating the docs

plt.show()



"""
SunPy Sphinx Theme.
"""

import json
import os
from functools import partial
from pathlib import Path
from textwrap import dedent, indent
from urllib.parse import urljoin

from pydata_sphinx_theme import utils
from sphinx.application import Sphinx

__all__ = ["ON_RTD", "PNG_ICON", "SVG_ICON", "get_html_theme_path"]


def get_html_theme_path():
    """
    Return list of HTML theme paths.
    """
    parent = Path(__file__).parent.resolve()
    return parent / "theme" / "sunpy"


def default_navbar():
    return [
        (
            "About",
            [
                ("Our Mission", "about/", 2),
                ("SunPy Project", "about/project/", 2),
                ("Presentations", "about/presentations/", 2),
                ("Meetings", "about/meetings/", 2),
                ("Code of Conduct", "coc/", 2),
            ],
        ),
        (
            "Documentation",
            [
                # Core goes first always
                ("sunpy", "https://docs.sunpy.org/", 3),
                # Other affiliated packages are in alphabetical order
                ("aiapy", "https://aiapy.readthedocs.io/", 3),
                ("dkist", "https://docs.dkist.nso.edu/projects/python-tools", 3),
                ("drms", "https://docs.sunpy.org/projects/drms/", 3),
                ("irispy", "https://irispy.readthedocs.io/", 3),
                ("ndcube", "https://docs.sunpy.org/projects/ndcube/", 3),
                ("roentgen", "https://roentgen.readthedocs.io/", 3),
                ("scope", "https://statistical-confidence-of-oscillatory-processes-with-emd.readthedocs.io", 3),
                ("solarmach", "https://solarmach.readthedocs.io/en/stable/", 3),
                ("sunkit-image", "https://docs.sunpy.org/projects/sunkit-image/", 3),
                ("sunkit-instruments ", "https://docs.sunpy.org/projects/sunkit-instruments/", 3),
                ("sunkit-magex", "https://docs.sunpy.org/projects/sunkit-magex/", 3),
                ("sunkit-pyvista", "https://docs.sunpy.org/projects/sunkit-pyvista/", 3),
                ("sunpy-soar", "https://docs.sunpy.org/projects/soar/", 3),
                ("sunraster", "https://docs.sunpy.org/projects/sunraster/", 3),
                ("xrtpy", "https://xrtpy.readthedocs.io/", 3),
                # Provisional packages submenu
                (
                    "Provisional",
                    [
                        ("pyflct", "https://pyflct.readthedocs.io/", 3),
                        ("radiospectra", "https://docs.sunpy.org/projects/radiospectra/", 3),
                    ],
                ),
                # Tools submenu
                (
                    "Tools",
                    [
                        ("ablog", "https://ablog.readthedocs.io/en/stable/", 3),
                        ("mpl-animators", "https://docs.sunpy.org/projects/mpl-animators/", 3),
                        ("streamtracer", "https://docs.sunpy.org/projects/streamtracer/", 3),
                    ],
                ),
            ],
        ),
        ("Packages", "affiliated/", 2),
        ("Get Help", "help/", 2),
        ("Contribute", "contribute/", 2),
        ("Blog", "blog/", 2),
        ("Cite SunPy", "https://docs.sunpy.org/en/stable/citation.html", 3),
    ]


def update_config(app) -> None:
    """
    Update config with new default values and handle deprecated keys.
    """
    # By the time `builder-inited` happens, `app.builder.theme_options` already exists.
    # At this point, modifying app.config.html_theme_options will NOT update the
    # page's HTML context (e.g. in jinja, `theme_keyword`).
    # To do this, you must manually modify `app.builder.theme_options`.
    theme_options = utils.get_theme_options_dict(app)

    if theme_options.get("sst_logo") and not isinstance(theme_options["sst_logo"], dict):
        sst_logo = str(theme_options["sst_logo"])
        theme_options["sst_logo"] = {"light": sst_logo, "dark": sst_logo}

    theme_options["sst_is_root"] = bool(theme_options.get("sst_is_root", False))

    if not theme_options.get("navbar_links"):
        theme_options["navbar_links"] = default_navbar()

    if not theme_options.get("footer_links", False):
        theme_options["footer_links"] = [
            ("Code", "https://github.com/sunpy", 3),
            ("Forum", "https://community.openastronomy.org/c/sunpy", 3),
            ("Chat", "https://openastronomy.element.io/#/room/#sunpy:openastronomy.org", 3),
        ]

    # TODO: This is nasty
    # Set the default value of show_source to False unless it's specified in the user config
    if "html_show_sourcelink" not in app.config._raw_config:  # NOQA: SLF001
        app.config.html_show_sourcelink = False

    # Set the logo to the sunpy logo unless it's overridden in the user config
    if "html_logo" not in app.config._raw_config:  # NOQA: SLF001
        app.config.html_logo = str(get_html_theme_path() / "static" / "img" / "sunpy_icon.svg")


def sst_pathto(context, document, relative_to=0):
    """
    This is a modified version of the built-in ``pathto()`` function.

    The default version when called with one argument returns the URL to a
    sphinx document, when specified with a 1 as the second argument it returns
    the path to a file relative to the root of the generated output.

    This version has 4 modes:
    * ``sst_pathto(document)`` - The same as ``pathto(document)``
    * ``sst_pathto(document, 1)`` - The same as ``pathto(document, 1)``
    * ``sst_pathto(document, 2)`` - A URL relative to ``sst_site_root`` will be returned when ``sst_is_root`` is ``False`` and the equivalent of specifying 1 as the second argument when it is ``True``.
    * ``sst_pathto(document, 3)`` - Do nothing return ``document`` unmodified.
    """
    if relative_to == 0:
        return context["pathto"](document)
    elif relative_to == 1:  # NOQA: RET505
        return context["pathto"](document, 1)
    elif relative_to == 2:
        if context.get("theme_sst_is_root", False):
            return context["pathto"](document, 1)
        return urljoin(context["theme_sst_site_root"], document)
    elif relative_to == 3:
        return document
    else:
        msg = "The third element of a link tuple must be 1, 2 or 3"
        raise ValueError(msg)


def update_html_context(app: Sphinx, pagename: str, templatename: str, context, doctree) -> None:  # NOQA: ARG001
    """
    Set extra things to use in jinja templates.
    """
    context["favicon_url"] = context.get("favicon_url", None) or "_static/img/sunpy_icon.svg"
    context["sst_pathto"] = partial(sst_pathto, context)


def generate_search_config(app):
    """
    This function parses the config for the "Documentation" section of the theme config.
    """
    theme_config = utils.get_theme_options_dict(app)
    search_projects = theme_config.get("rtd_search_projects", None)
    if search_projects is None:
        navbar_links = theme_config["navbar_links"]
        doc_links = next(section[1] for section in navbar_links if section[0] == "Documentation")

        def filter_doc_links(links):
            out_links = []
            for link in links:
                if isinstance(link[1], list):
                    out_links += filter_doc_links(link[1])
                elif isinstance(link[1], str) and link[1].startswith("http"):
                    out_links.append({"name": link[0], "link": link[1]})
                else:
                    err = f"Unable to parse {link} in the nav tree. Try setting search_projects explicitly or fixing navbar_links."
                    raise ValueError(err)
            return out_links

        search_projects = filter_doc_links(doc_links)

    load_more_label = theme_config.get("rtd_search_load_more_label", "Load more results")
    no_results_label = theme_config.get("rtd_search_no_results_label", "There are no results for this search")
    script = dedent(f"""
        const set_search_config = {{
          "no-results":{{
            "label": "{no_results_label}"
          }},
          "load-more":{{
            "label": "{load_more_label}",
            "class": "btn sd-btn sd-bg-primary sd-bg-text-primary"
          }},
          "projects":{indent(json.dumps(search_projects, indent=2), " " * 10, predicate=lambda line: line.strip() != "[")}
        }};
    """)
    app.add_js_file(None, body=script)


def setup(app: Sphinx):
    # Register theme
    theme_dir = get_html_theme_path()
    app.add_html_theme("sunpy", theme_dir)
    app.add_css_file("sunpy_style.css", priority=600)
    app.connect("builder-inited", update_config, priority=100)
    app.connect("builder-inited", generate_search_config, priority=500)
    app.connect("html-page-context", update_html_context)
    # Conditionally include goat counter js
    # We can't do this in update_config as that causes the scripts to be duplicated.
    # Also in here none of the theme defaults have be applied by `update_config`
    # TODO: Improve this mess
    theme_options = utils.get_theme_options_dict(app)
    # We want to default to the sunpy goat counter only if the sst_site_root is sunpy.org
    root_domain = theme_options.get("sst_site_root", "https://sunpy.org")
    sunpy_goat_url = "https://sunpy.goatcounter.com/count"
    default_goat_url = sunpy_goat_url if root_domain == "https://sunpy.org" else None
    if primary_goat_url := theme_options.get("goatcounter_analytics_url", default_goat_url):
        root_domain = root_domain.removeprefix("https://").removeprefix("http://")
        default_endpoint = theme_options.get("goatcounter_non_domain_endpoint", False)
        if default_endpoint is False:
            default_endpoint = ""
        app.add_js_file(
            None,
            body=f"""
            var endpoint = '{default_endpoint}';
            if (location.hostname.endsWith('{root_domain}')) {{
                endpoint = '{primary_goat_url}'
            }}

            window.goatcounter = {{
                endpoint: endpoint,
                path: function(p) {{ return location.host + p }}
            }}
            """,
        )
        app.add_js_file(
            "https://gc.zgo.at/count.js",
            loading_method="async",
        )
        app.add_js_file(
            "js/submenu-concertina-toggle.js",
            loading_method="async",
        )

        if theme_options.get("rtd_search", True):
            # Add project-wide search
            app.add_css_file("css/rtd_enhanced_search.css")
            app.add_js_file(
                "js/rtd_enhanced_search.js",
                loading_method="async",
            )

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


_sunpy_static_path = get_html_theme_path() / "static"
ON_RTD = os.environ.get("READTHEDOCS", "False") == "True"
SVG_ICON = _sunpy_static_path / "img" / "sunpy_icon.svg"
PNG_ICON = _sunpy_static_path / "img" / "sunpy_icon_128x128.png"



"""
This config file is kept for backwards compatibility, almost all the config has
now been moved into the theme itself or can be imported from the main namespace
of the theme.
"""

from sphinx.util import logging

from sunpy_sphinx_theme import ON_RTD as on_rtd  # noqa: N811
from sunpy_sphinx_theme import PNG_ICON as png_icon  # noqa: N811
from sunpy_sphinx_theme import SVG_ICON as svg_icon  # noqa: N811
from sunpy_sphinx_theme import get_html_theme_path

_log = logging.getLogger(__name__)

_log.warning(
    "Using sunpy_sphinx_theme.conf is deprecated, please migrate to importing things you need directly from sunpy_sphinx_theme."
)

__all__ = [
    "html_static_path",
    "html_theme",
    "html_theme_path",
    "on_rtd",
    "png_icon",
    "svg_icon",
]

html_theme = "sunpy"
html_theme_options = {}

html_theme_path = [str(get_html_theme_path())]
html_static_path = [str(get_html_theme_path() / "static")]



"""
Run all builds for all active versions of all SunPy and SunPy subprojects on
RTD.
"""

import os

import requests

ORG_PROJECT = "sunpy"
BASE_URL = "https://readthedocs.org/api/v3/projects/"
HEADERS = {"Authorization": f"Token {os.environ['RTD_AUTH_TOKEN']}"}
# RTD removes the full stop from the project slug
WEBSITE_PROJECT = "sunpyorg"


def get_active_versions(project):
    r = requests.get(f"{BASE_URL}{project}/versions", headers=HEADERS, params={"active": True})
    if not r.ok:
        print(f"Failed to get versions for {project}: {r}")
        return []
    r = r.json()
    if "results" not in r:
        print(project)
        print(r)
        return []
    results = r["results"]
    return [res["slug"] for res in results]


def get_all_subprojects(base_project):
    r = requests.get(f"{BASE_URL}{base_project}/subprojects", headers=HEADERS)
    r = r.json()
    results = r["results"]
    return [res["child"]["slug"] for res in results]


def rebuild_all_versions_for_project(project):
    slugs = get_active_versions(project)
    for slug in slugs:
        r = requests.post(f"{BASE_URL}{project}/versions/{slug}/builds/", headers=HEADERS)
        if r.status_code != 202:
            print(f"{slug} failed to build with: {r}")


if __name__ == "__main__":
    projects = [ORG_PROJECT, *get_all_subprojects(ORG_PROJECT), WEBSITE_PROJECT]
    for project in projects:
        rebuild_all_versions_for_project(project)



"""
This module provides a collection of functions.
"""

from sunpy.util.decorators import deprecated

__all__ = [
    "Animal",
    "a_really_long_function_name_just_to_see_what_happens",
    "function",
]


@deprecated("1.0", alternative="sunpy.net.Fido")
class Animal:
    """
    A deprecated class used to represent an Animal.

    .. note::

        * This is a note.

    .. warning::

        * This is a warning.

    Attributes
    ----------
    name : `str`
        The name of the animal.
    sound : `str`
        The sound that the animal makes.
    num_legs : int
        The number of legs the animal has (default 4).

    Examples
    --------
    >>> import datetime
    >>> datetime.datetime(2019, 8, 16, 22, 46, 37, 856437)
    datetime.datetime(2019, 8, 16, 22, 46, 37, 856437)

    References
    ----------
    * https://realpython.com/documenting-python-code/
    """

    says_str = "A {name} says {sound}"

    def __init__(self, name, sound, num_legs=5) -> None:
        """
        Parameters
        ----------
        name : `str`
            The name of the animal.
        sound : `str`
            The sound the animal makes.
        num_legs : int, optional
            The number of legs the animal (default is 5).
        """
        self.name = name
        self.sound = sound
        self.num_legs = num_legs

    def says(self, sound=None) -> None:
        """
        Prints what the animals name is and what sound it makes.

        If the argument `sound` isn't passed in, the default animal
        sound is used.

        Parameters
        ----------
        sound : `str`, optional
            The sound the animal makes (default is `None`),

        Raises
        ------
        `NotImplementedError`
            If no sound is set for the animal or passed in as a
            parameter.

        References
        ----------
        * `A URL. <www.sunpy.org>`__
        """
        if self.sound is None and sound is None:
            msg = "Silent Animals are not supported!"
            raise NotImplementedError(msg)


def function() -> None:
    """
    Prints what the animals name is and what sound it makes.

    If the argument `sound` isn't passed in, the default animal
    sound is used.

    .. note::

        * This is a note.

    .. warning::

        * This is a warning.

    Parameters
    ----------
    sound : `str`, optional
        The sound the animal makes (default is `None`).

    Returns
    -------
    `list`
        A list.

    Raises
    ------
    `NotImplementedError`
        If no sound is set for the animal or passed in as a
        parameter.

    Examples
    --------
    >>> import datetime
    >>> datetime.datetime(2019, 8, 16, 22, 46, 37, 856437)
    datetime.datetime(2019, 8, 16, 22, 46, 37, 856437)

    References
    ----------
    * `A URL. <www.sunpy.org>`__
    """


def a_really_long_function_name_just_to_see_what_happens() -> None:
    """
    Prints what the animals name is and what sound it makes.

    If the argument `sound` isn't passed in, the default animal
    sound is used.

    .. note::

        * This is a note.

    .. warning::

        * This is a warning.

    Parameters
    ----------
    sound : `str`, optional
        The sound the animal makes (default is `None`).

    Returns
    -------
    `list`
        A list.

    Raises
    ------
    `NotImplementedError`
        If no sound is set for the animal or passed in as a
        parameter.

    Examples
    --------
    >>> import datetime
    >>> datetime.datetime(2019, 8, 16, 22, 46, 37, 856437)
    datetime.datetime(2019, 8, 16, 22, 46, 37, 856437)

    References
    ----------
    * `A URL. <www.sunpy.org>`__
    """
